from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from pydantic import Field

from kubernetes_dynamic.models.common import get_default

from .resource_item import ResourceItem

if TYPE_CHECKING:
    from .all import V1IngressSpec, V1IngressStatus


class V1Ingress(ResourceItem):
    spec: V1IngressSpec = Field(default_factory=lambda: get_default("V1IngressSpec"))
    status: V1IngressStatus = Field(default_factory=lambda: get_default("V1IngressStatus"))

    @classmethod
    def get_default_host(cls) -> Optional[str]:
        """Get default ingress host."""
        items = cls.default_client().ingresses.get()
        if not items:
            return None
        return items[0].url

    @property
    def url(self):
        """Get URL."""
        return f"https://{self.spec.rules[0].host}"



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from kubernetes_dynamic.formatters import format_selector

from .common import ItemList, V1ObjectMeta, get_default
from .resource_item import ResourceItem

if TYPE_CHECKING:
    from .all import V1ReplicaSetSpec, V1ReplicaSetStatus
    from .pod import V1Pod


class V1ReplicaSet(ResourceItem):
    metadata: V1ObjectMeta = Field(default_factory=lambda: V1ObjectMeta())
    spec: V1ReplicaSetSpec = Field(default_factory=lambda: get_default("V1ReplicaSetSpec"))
    status: V1ReplicaSetStatus = Field(default_factory=lambda: get_default("V1ReplicaSetStatus"))

    def get_pods(self) -> ItemList[V1Pod]:
        label_selector = self.spec.selector.matchLabels
        return self._client.pods.get(label_selector=format_selector(label_selector))



__all__ = [
    "AdmissionregistrationV1ServiceReference",
    "AdmissionregistrationV1WebhookClientConfig",
    "ApiextensionsV1ServiceReference",
    "ApiextensionsV1WebhookClientConfig",
    "ApiregistrationV1ServiceReference",
    "AuthenticationV1TokenRequest",
    "CoreV1EndpointPort",
    "CoreV1Event",
    "CoreV1EventList",
    "CoreV1EventSeries",
    "DiscoveryV1EndpointPort",
    "EventsV1Event",
    "EventsV1EventList",
    "EventsV1EventSeries",
    "StorageV1TokenRequest",
    "V1APIGroup",
    "V1APIGroupList",
    "V1APIResource",
    "V1APIResourceList",
    "V1APIService",
    "V1APIServiceCondition",
    "V1APIServiceList",
    "V1APIServiceSpec",
    "V1APIServiceStatus",
    "V1APIVersions",
    "V1AWSElasticBlockStoreVolumeSource",
    "V1Affinity",
    "V1AggregationRule",
    "V1AttachedVolume",
    "V1AzureDiskVolumeSource",
    "V1AzureFilePersistentVolumeSource",
    "V1AzureFileVolumeSource",
    "V1Binding",
    "V1BoundObjectReference",
    "V1CSIDriver",
    "V1CSIDriverList",
    "V1CSIDriverSpec",
    "V1CSINode",
    "V1CSINodeDriver",
    "V1CSINodeList",
    "V1CSINodeSpec",
    "V1CSIPersistentVolumeSource",
    "V1CSIStorageCapacity",
    "V1CSIStorageCapacityList",
    "V1CSIVolumeSource",
    "V1Capabilities",
    "V1CephFSPersistentVolumeSource",
    "V1CephFSVolumeSource",
    "V1CertificateSigningRequest",
    "V1CertificateSigningRequestCondition",
    "V1CertificateSigningRequestList",
    "V1CertificateSigningRequestSpec",
    "V1CertificateSigningRequestStatus",
    "V1CinderPersistentVolumeSource",
    "V1CinderVolumeSource",
    "V1ClaimSource",
    "V1ClientIPConfig",
    "V1ClusterRole",
    "V1ClusterRoleBinding",
    "V1ClusterRoleBindingList",
    "V1ClusterRoleList",
    "V1ComponentCondition",
    "V1ComponentStatus",
    "V1ComponentStatusList",
    "V1Condition",
    "V1ConfigMap",
    "V1ConfigMapEnvSource",
    "V1ConfigMapKeySelector",
    "V1ConfigMapList",
    "V1ConfigMapNodeConfigSource",
    "V1ConfigMapProjection",
    "V1ConfigMapVolumeSource",
    "V1Container",
    "V1ContainerImage",
    "V1ContainerPort",
    "V1ContainerState",
    "V1ContainerStateRunning",
    "V1ContainerStateTerminated",
    "V1ContainerStateWaiting",
    "V1ContainerStatus",
    "V1ControllerRevision",
    "V1ControllerRevisionList",
    "V1CronJob",
    "V1CronJobList",
    "V1CronJobSpec",
    "V1CronJobStatus",
    "V1CrossVersionObjectReference",
    "V1CustomResourceColumnDefinition",
    "V1CustomResourceConversion",
    "V1CustomResourceDefinition",
    "V1CustomResourceDefinitionCondition",
    "V1CustomResourceDefinitionList",
    "V1CustomResourceDefinitionNames",
    "V1CustomResourceDefinitionSpec",
    "V1CustomResourceDefinitionStatus",
    "V1CustomResourceDefinitionVersion",
    "V1CustomResourceSubresourceScale",
    "V1CustomResourceSubresources",
    "V1CustomResourceValidation",
    "V1DaemonEndpoint",
    "V1DaemonSet",
    "V1DaemonSetCondition",
    "V1DaemonSetList",
    "V1DaemonSetSpec",
    "V1DaemonSetStatus",
    "V1DaemonSetUpdateStrategy",
    "V1DeleteOptions",
    "V1Deployment",
    "V1DeploymentCondition",
    "V1DeploymentList",
    "V1DeploymentSpec",
    "V1DeploymentStatus",
    "V1DeploymentStrategy",
    "V1DownwardAPIProjection",
    "V1DownwardAPIVolumeFile",
    "V1DownwardAPIVolumeSource",
    "V1EmptyDirVolumeSource",
    "V1Endpoint",
    "V1EndpointAddress",
    "V1EndpointConditions",
    "V1EndpointHints",
    "V1EndpointSlice",
    "V1EndpointSliceList",
    "V1EndpointSubset",
    "V1Endpoints",
    "V1EndpointsList",
    "V1EnvFromSource",
    "V1EnvVar",
    "V1EnvVarSource",
    "V1EphemeralContainer",
    "V1EphemeralVolumeSource",
    "V1EventSource",
    "V1Eviction",
    "V1ExecAction",
    "V1ExternalDocumentation",
    "V1FCVolumeSource",
    "V1FlexPersistentVolumeSource",
    "V1FlexVolumeSource",
    "V1FlockerVolumeSource",
    "V1ForZone",
    "V1GCEPersistentDiskVolumeSource",
    "V1GRPCAction",
    "V1GitRepoVolumeSource",
    "V1GlusterfsPersistentVolumeSource",
    "V1GlusterfsVolumeSource",
    "V1GroupVersionForDiscovery",
    "V1HTTPGetAction",
    "V1HTTPHeader",
    "V1HTTPIngressPath",
    "V1HTTPIngressRuleValue",
    "V1HorizontalPodAutoscaler",
    "V1HorizontalPodAutoscalerList",
    "V1HorizontalPodAutoscalerSpec",
    "V1HorizontalPodAutoscalerStatus",
    "V1HostAlias",
    "V1HostPathVolumeSource",
    "V1IPBlock",
    "V1ISCSIPersistentVolumeSource",
    "V1ISCSIVolumeSource",
    "V1Ingress",
    "V1IngressBackend",
    "V1IngressClass",
    "V1IngressClassList",
    "V1IngressClassParametersReference",
    "V1IngressClassSpec",
    "V1IngressList",
    "V1IngressLoadBalancerIngress",
    "V1IngressLoadBalancerStatus",
    "V1IngressPortStatus",
    "V1IngressRule",
    "V1IngressServiceBackend",
    "V1IngressSpec",
    "V1IngressStatus",
    "V1IngressTLS",
    "V1JSONSchemaProps",
    "V1Job",
    "V1JobCondition",
    "V1JobList",
    "V1JobSpec",
    "V1JobStatus",
    "V1JobTemplateSpec",
    "V1KeyToPath",
    "V1LabelSelector",
    "V1LabelSelectorRequirement",
    "V1Lease",
    "V1LeaseList",
    "V1LeaseSpec",
    "V1Lifecycle",
    "V1LifecycleHandler",
    "V1LimitRange",
    "V1LimitRangeItem",
    "V1LimitRangeList",
    "V1LimitRangeSpec",
    "V1ListMeta",
    "V1LoadBalancerIngress",
    "V1LoadBalancerStatus",
    "V1LocalObjectReference",
    "V1LocalSubjectAccessReview",
    "V1LocalVolumeSource",
    "V1ManagedFieldsEntry",
    "V1MutatingWebhook",
    "V1MutatingWebhookConfiguration",
    "V1MutatingWebhookConfigurationList",
    "V1NFSVolumeSource",
    "V1Namespace",
    "V1NamespaceCondition",
    "V1NamespaceList",
    "V1NamespaceSpec",
    "V1NamespaceStatus",
    "V1NetworkPolicy",
    "V1NetworkPolicyEgressRule",
    "V1NetworkPolicyIngressRule",
    "V1NetworkPolicyList",
    "V1NetworkPolicyPeer",
    "V1NetworkPolicyPort",
    "V1NetworkPolicySpec",
    "V1NetworkPolicyStatus",
    "V1Node",
    "V1NodeAddress",
    "V1NodeAffinity",
    "V1NodeCondition",
    "V1NodeConfigSource",
    "V1NodeConfigStatus",
    "V1NodeDaemonEndpoints",
    "V1NodeList",
    "V1NodeSelector",
    "V1NodeSelectorRequirement",
    "V1NodeSelectorTerm",
    "V1NodeSpec",
    "V1NodeStatus",
    "V1NodeSystemInfo",
    "V1NonResourceAttributes",
    "V1NonResourceRule",
    "V1ObjectFieldSelector",
    "V1ObjectMeta",
    "V1ObjectReference",
    "V1Overhead",
    "V1OwnerReference",
    "V1PersistentVolume",
    "V1PersistentVolumeClaim",
    "V1PersistentVolumeClaimCondition",
    "V1PersistentVolumeClaimList",
    "V1PersistentVolumeClaimSpec",
    "V1PersistentVolumeClaimStatus",
    "V1PersistentVolumeClaimTemplate",
    "V1PersistentVolumeClaimVolumeSource",
    "V1PersistentVolumeList",
    "V1PersistentVolumeSpec",
    "V1PersistentVolumeStatus",
    "V1PhotonPersistentDiskVolumeSource",
    "V1Pod",
    "V1PodAffinity",
    "V1PodAffinityTerm",
    "V1PodAntiAffinity",
    "V1PodCondition",
    "V1PodDNSConfig",
    "V1PodDNSConfigOption",
    "V1PodDisruptionBudget",
    "V1PodDisruptionBudgetList",
    "V1PodDisruptionBudgetSpec",
    "V1PodDisruptionBudgetStatus",
    "V1PodFailurePolicy",
    "V1PodFailurePolicyOnExitCodesRequirement",
    "V1PodFailurePolicyOnPodConditionsPattern",
    "V1PodFailurePolicyRule",
    "V1PodIP",
    "V1PodList",
    "V1PodOS",
    "V1PodReadinessGate",
    "V1PodResourceClaim",
    "V1PodSchedulingGate",
    "V1PodSecurityContext",
    "V1PodSpec",
    "V1PodStatus",
    "V1PodTemplate",
    "V1PodTemplateList",
    "V1PodTemplateSpec",
    "V1PolicyRule",
    "V1PortStatus",
    "V1PortworxVolumeSource",
    "V1Preconditions",
    "V1PreferredSchedulingTerm",
    "V1PriorityClass",
    "V1PriorityClassList",
    "V1Probe",
    "V1ProjectedVolumeSource",
    "V1QuobyteVolumeSource",
    "V1RBDPersistentVolumeSource",
    "V1RBDVolumeSource",
    "V1ReplicaSet",
    "V1ReplicaSetCondition",
    "V1ReplicaSetList",
    "V1ReplicaSetSpec",
    "V1ReplicaSetStatus",
    "V1ReplicationController",
    "V1ReplicationControllerCondition",
    "V1ReplicationControllerList",
    "V1ReplicationControllerSpec",
    "V1ReplicationControllerStatus",
    "V1ResourceAttributes",
    "V1ResourceClaim",
    "V1ResourceFieldSelector",
    "V1ResourceQuota",
    "V1ResourceQuotaList",
    "V1ResourceQuotaSpec",
    "V1ResourceQuotaStatus",
    "V1ResourceRequirements",
    "V1ResourceRule",
    "V1Role",
    "V1RoleBinding",
    "V1RoleBindingList",
    "V1RoleList",
    "V1RoleRef",
    "V1RollingUpdateDaemonSet",
    "V1RollingUpdateDeployment",
    "V1RollingUpdateStatefulSetStrategy",
    "V1RuleWithOperations",
    "V1RuntimeClass",
    "V1RuntimeClassList",
    "V1SELinuxOptions",
    "V1Scale",
    "V1ScaleIOPersistentVolumeSource",
    "V1ScaleIOVolumeSource",
    "V1ScaleSpec",
    "V1ScaleStatus",
    "V1Scheduling",
    "V1ScopeSelector",
    "V1ScopedResourceSelectorRequirement",
    "V1SeccompProfile",
    "V1Secret",
    "V1SecretEnvSource",
    "V1SecretKeySelector",
    "V1SecretList",
    "V1SecretProjection",
    "V1SecretReference",
    "V1SecretVolumeSource",
    "V1SecurityContext",
    "V1SelfSubjectAccessReview",
    "V1SelfSubjectAccessReviewSpec",
    "V1SelfSubjectRulesReview",
    "V1SelfSubjectRulesReviewSpec",
    "V1ServerAddressByClientCIDR",
    "V1Service",
    "V1ServiceAccount",
    "V1ServiceAccountList",
    "V1ServiceAccountTokenProjection",
    "V1ServiceBackendPort",
    "V1ServiceList",
    "V1ServicePort",
    "V1ServiceSpec",
    "V1ServiceStatus",
    "V1SessionAffinityConfig",
    "V1StatefulSet",
    "V1StatefulSetCondition",
    "V1StatefulSetList",
    "V1StatefulSetOrdinals",
    "V1StatefulSetPersistentVolumeClaimRetentionPolicy",
    "V1StatefulSetSpec",
    "V1StatefulSetStatus",
    "V1StatefulSetUpdateStrategy",
    "V1Status",
    "V1StatusCause",
    "V1StatusDetails",
    "V1StorageClass",
    "V1StorageClassList",
    "V1StorageOSPersistentVolumeSource",
    "V1StorageOSVolumeSource",
    "V1Subject",
    "V1SubjectAccessReview",
    "V1SubjectAccessReviewSpec",
    "V1SubjectAccessReviewStatus",
    "V1SubjectRulesReviewStatus",
    "V1Sysctl",
    "V1TCPSocketAction",
    "V1Taint",
    "V1TokenRequestSpec",
    "V1TokenRequestStatus",
    "V1TokenReview",
    "V1TokenReviewSpec",
    "V1TokenReviewStatus",
    "V1Toleration",
    "V1TopologySelectorLabelRequirement",
    "V1TopologySelectorTerm",
    "V1TopologySpreadConstraint",
    "V1TypedLocalObjectReference",
    "V1TypedObjectReference",
    "V1UncountedTerminatedPods",
    "V1UserInfo",
    "V1ValidatingWebhook",
    "V1ValidatingWebhookConfiguration",
    "V1ValidatingWebhookConfigurationList",
    "V1Volume",
    "V1VolumeAttachment",
    "V1VolumeAttachmentList",
    "V1VolumeAttachmentSource",
    "V1VolumeAttachmentSpec",
    "V1VolumeAttachmentStatus",
    "V1VolumeDevice",
    "V1VolumeError",
    "V1VolumeMount",
    "V1VolumeNodeAffinity",
    "V1VolumeNodeResources",
    "V1VolumeProjection",
    "V1VsphereVirtualDiskVolumeSource",
    "V1WebhookConversion",
    "V1WeightedPodAffinityTerm",
    "V1WindowsSecurityContextOptions",
    "V1alpha1AllocationResult",
    "V1alpha1ClusterCIDR",
    "V1alpha1ClusterCIDRList",
    "V1alpha1ClusterCIDRSpec",
    "V1alpha1MatchResources",
    "V1alpha1NamedRuleWithOperations",
    "V1alpha1ParamKind",
    "V1alpha1ParamRef",
    "V1alpha1PodScheduling",
    "V1alpha1PodSchedulingList",
    "V1alpha1PodSchedulingSpec",
    "V1alpha1PodSchedulingStatus",
    "V1alpha1ResourceClaim",
    "V1alpha1ResourceClaimConsumerReference",
    "V1alpha1ResourceClaimList",
    "V1alpha1ResourceClaimParametersReference",
    "V1alpha1ResourceClaimSchedulingStatus",
    "V1alpha1ResourceClaimSpec",
    "V1alpha1ResourceClaimStatus",
    "V1alpha1ResourceClaimTemplate",
    "V1alpha1ResourceClaimTemplateList",
    "V1alpha1ResourceClaimTemplateSpec",
    "V1alpha1ResourceClass",
    "V1alpha1ResourceClassList",
    "V1alpha1ResourceClassParametersReference",
    "V1alpha1SelfSubjectReview",
    "V1alpha1SelfSubjectReviewStatus",
    "V1alpha1ServerStorageVersion",
    "V1alpha1StorageVersion",
    "V1alpha1StorageVersionCondition",
    "V1alpha1StorageVersionList",
    "V1alpha1StorageVersionStatus",
    "V1alpha1ValidatingAdmissionPolicy",
    "V1alpha1ValidatingAdmissionPolicyBinding",
    "V1alpha1ValidatingAdmissionPolicyBindingList",
    "V1alpha1ValidatingAdmissionPolicyBindingSpec",
    "V1alpha1ValidatingAdmissionPolicyList",
    "V1alpha1ValidatingAdmissionPolicySpec",
    "V1alpha1Validation",
    "V1beta1CSIStorageCapacity",
    "V1beta1CSIStorageCapacityList",
    "V1beta2FlowDistinguisherMethod",
    "V1beta2FlowSchema",
    "V1beta2FlowSchemaCondition",
    "V1beta2FlowSchemaList",
    "V1beta2FlowSchemaSpec",
    "V1beta2FlowSchemaStatus",
    "V1beta2GroupSubject",
    "V1beta2LimitResponse",
    "V1beta2LimitedPriorityLevelConfiguration",
    "V1beta2NonResourcePolicyRule",
    "V1beta2PolicyRulesWithSubjects",
    "V1beta2PriorityLevelConfiguration",
    "V1beta2PriorityLevelConfigurationCondition",
    "V1beta2PriorityLevelConfigurationList",
    "V1beta2PriorityLevelConfigurationReference",
    "V1beta2PriorityLevelConfigurationSpec",
    "V1beta2PriorityLevelConfigurationStatus",
    "V1beta2QueuingConfiguration",
    "V1beta2ResourcePolicyRule",
    "V1beta2ServiceAccountSubject",
    "V1beta2Subject",
    "V1beta2UserSubject",
    "V1beta3FlowDistinguisherMethod",
    "V1beta3FlowSchema",
    "V1beta3FlowSchemaCondition",
    "V1beta3FlowSchemaList",
    "V1beta3FlowSchemaSpec",
    "V1beta3FlowSchemaStatus",
    "V1beta3GroupSubject",
    "V1beta3LimitResponse",
    "V1beta3LimitedPriorityLevelConfiguration",
    "V1beta3NonResourcePolicyRule",
    "V1beta3PolicyRulesWithSubjects",
    "V1beta3PriorityLevelConfiguration",
    "V1beta3PriorityLevelConfigurationCondition",
    "V1beta3PriorityLevelConfigurationList",
    "V1beta3PriorityLevelConfigurationReference",
    "V1beta3PriorityLevelConfigurationSpec",
    "V1beta3PriorityLevelConfigurationStatus",
    "V1beta3QueuingConfiguration",
    "V1beta3ResourcePolicyRule",
    "V1beta3ServiceAccountSubject",
    "V1beta3Subject",
    "V1beta3UserSubject",
    "V2ContainerResourceMetricSource",
    "V2ContainerResourceMetricStatus",
    "V2CrossVersionObjectReference",
    "V2ExternalMetricSource",
    "V2ExternalMetricStatus",
    "V2HPAScalingPolicy",
    "V2HPAScalingRules",
    "V2HorizontalPodAutoscaler",
    "V2HorizontalPodAutoscalerBehavior",
    "V2HorizontalPodAutoscalerCondition",
    "V2HorizontalPodAutoscalerList",
    "V2HorizontalPodAutoscalerSpec",
    "V2HorizontalPodAutoscalerStatus",
    "V2MetricIdentifier",
    "V2MetricSpec",
    "V2MetricStatus",
    "V2MetricTarget",
    "V2MetricValueStatus",
    "V2ObjectMetricSource",
    "V2ObjectMetricStatus",
    "V2PodsMetricSource",
    "V2PodsMetricStatus",
    "V2ResourceMetricSource",
    "V2ResourceMetricStatus",
]

from kubernetes_dynamic.models.all import (
    AdmissionregistrationV1ServiceReference,
    AdmissionregistrationV1WebhookClientConfig,
    ApiextensionsV1ServiceReference,
    ApiextensionsV1WebhookClientConfig,
    ApiregistrationV1ServiceReference,
    AuthenticationV1TokenRequest,
    CoreV1EndpointPort,
    CoreV1Event,
    CoreV1EventList,
    CoreV1EventSeries,
    DiscoveryV1EndpointPort,
    EventsV1Event,
    EventsV1EventList,
    EventsV1EventSeries,
    StorageV1TokenRequest,
    V1Affinity,
    V1AggregationRule,
    V1alpha1AllocationResult,
    V1alpha1ClusterCIDR,
    V1alpha1ClusterCIDRList,
    V1alpha1ClusterCIDRSpec,
    V1alpha1MatchResources,
    V1alpha1NamedRuleWithOperations,
    V1alpha1ParamKind,
    V1alpha1ParamRef,
    V1alpha1PodScheduling,
    V1alpha1PodSchedulingList,
    V1alpha1PodSchedulingSpec,
    V1alpha1PodSchedulingStatus,
    V1alpha1ResourceClaim,
    V1alpha1ResourceClaimConsumerReference,
    V1alpha1ResourceClaimList,
    V1alpha1ResourceClaimParametersReference,
    V1alpha1ResourceClaimSchedulingStatus,
    V1alpha1ResourceClaimSpec,
    V1alpha1ResourceClaimStatus,
    V1alpha1ResourceClaimTemplate,
    V1alpha1ResourceClaimTemplateList,
    V1alpha1ResourceClaimTemplateSpec,
    V1alpha1ResourceClass,
    V1alpha1ResourceClassList,
    V1alpha1ResourceClassParametersReference,
    V1alpha1SelfSubjectReview,
    V1alpha1SelfSubjectReviewStatus,
    V1alpha1ServerStorageVersion,
    V1alpha1StorageVersion,
    V1alpha1StorageVersionCondition,
    V1alpha1StorageVersionList,
    V1alpha1StorageVersionStatus,
    V1alpha1ValidatingAdmissionPolicy,
    V1alpha1ValidatingAdmissionPolicyBinding,
    V1alpha1ValidatingAdmissionPolicyBindingList,
    V1alpha1ValidatingAdmissionPolicyBindingSpec,
    V1alpha1ValidatingAdmissionPolicyList,
    V1alpha1ValidatingAdmissionPolicySpec,
    V1alpha1Validation,
    V1APIGroup,
    V1APIGroupList,
    V1APIResource,
    V1APIResourceList,
    V1APIService,
    V1APIServiceCondition,
    V1APIServiceList,
    V1APIServiceSpec,
    V1APIServiceStatus,
    V1APIVersions,
    V1AttachedVolume,
    V1AWSElasticBlockStoreVolumeSource,
    V1AzureDiskVolumeSource,
    V1AzureFilePersistentVolumeSource,
    V1AzureFileVolumeSource,
    V1beta1CSIStorageCapacity,
    V1beta1CSIStorageCapacityList,
    V1beta2FlowDistinguisherMethod,
    V1beta2FlowSchema,
    V1beta2FlowSchemaCondition,
    V1beta2FlowSchemaList,
    V1beta2FlowSchemaSpec,
    V1beta2FlowSchemaStatus,
    V1beta2GroupSubject,
    V1beta2LimitedPriorityLevelConfiguration,
    V1beta2LimitResponse,
    V1beta2NonResourcePolicyRule,
    V1beta2PolicyRulesWithSubjects,
    V1beta2PriorityLevelConfiguration,
    V1beta2PriorityLevelConfigurationCondition,
    V1beta2PriorityLevelConfigurationList,
    V1beta2PriorityLevelConfigurationReference,
    V1beta2PriorityLevelConfigurationSpec,
    V1beta2PriorityLevelConfigurationStatus,
    V1beta2QueuingConfiguration,
    V1beta2ResourcePolicyRule,
    V1beta2ServiceAccountSubject,
    V1beta2Subject,
    V1beta2UserSubject,
    V1beta3FlowDistinguisherMethod,
    V1beta3FlowSchema,
    V1beta3FlowSchemaCondition,
    V1beta3FlowSchemaList,
    V1beta3FlowSchemaSpec,
    V1beta3FlowSchemaStatus,
    V1beta3GroupSubject,
    V1beta3LimitedPriorityLevelConfiguration,
    V1beta3LimitResponse,
    V1beta3NonResourcePolicyRule,
    V1beta3PolicyRulesWithSubjects,
    V1beta3PriorityLevelConfiguration,
    V1beta3PriorityLevelConfigurationCondition,
    V1beta3PriorityLevelConfigurationList,
    V1beta3PriorityLevelConfigurationReference,
    V1beta3PriorityLevelConfigurationSpec,
    V1beta3PriorityLevelConfigurationStatus,
    V1beta3QueuingConfiguration,
    V1beta3ResourcePolicyRule,
    V1beta3ServiceAccountSubject,
    V1beta3Subject,
    V1beta3UserSubject,
    V1Binding,
    V1BoundObjectReference,
    V1Capabilities,
    V1CephFSPersistentVolumeSource,
    V1CephFSVolumeSource,
    V1CertificateSigningRequest,
    V1CertificateSigningRequestCondition,
    V1CertificateSigningRequestList,
    V1CertificateSigningRequestSpec,
    V1CertificateSigningRequestStatus,
    V1CinderPersistentVolumeSource,
    V1CinderVolumeSource,
    V1ClaimSource,
    V1ClientIPConfig,
    V1ClusterRole,
    V1ClusterRoleBinding,
    V1ClusterRoleBindingList,
    V1ClusterRoleList,
    V1ComponentCondition,
    V1ComponentStatus,
    V1ComponentStatusList,
    V1Condition,
    V1ConfigMap,
    V1ConfigMapEnvSource,
    V1ConfigMapKeySelector,
    V1ConfigMapList,
    V1ConfigMapNodeConfigSource,
    V1ConfigMapProjection,
    V1ConfigMapVolumeSource,
    V1Container,
    V1ContainerImage,
    V1ContainerPort,
    V1ContainerState,
    V1ContainerStateRunning,
    V1ContainerStateTerminated,
    V1ContainerStateWaiting,
    V1ContainerStatus,
    V1ControllerRevision,
    V1ControllerRevisionList,
    V1CronJob,
    V1CronJobList,
    V1CronJobSpec,
    V1CronJobStatus,
    V1CrossVersionObjectReference,
    V1CSIDriver,
    V1CSIDriverList,
    V1CSIDriverSpec,
    V1CSINode,
    V1CSINodeDriver,
    V1CSINodeList,
    V1CSINodeSpec,
    V1CSIPersistentVolumeSource,
    V1CSIStorageCapacity,
    V1CSIStorageCapacityList,
    V1CSIVolumeSource,
    V1CustomResourceColumnDefinition,
    V1CustomResourceConversion,
    V1CustomResourceDefinition,
    V1CustomResourceDefinitionCondition,
    V1CustomResourceDefinitionList,
    V1CustomResourceDefinitionNames,
    V1CustomResourceDefinitionSpec,
    V1CustomResourceDefinitionStatus,
    V1CustomResourceDefinitionVersion,
    V1CustomResourceSubresources,
    V1CustomResourceSubresourceScale,
    V1CustomResourceValidation,
    V1DaemonEndpoint,
    V1DaemonSet,
    V1DaemonSetCondition,
    V1DaemonSetList,
    V1DaemonSetSpec,
    V1DaemonSetStatus,
    V1DaemonSetUpdateStrategy,
    V1DeleteOptions,
    V1Deployment,
    V1DeploymentCondition,
    V1DeploymentList,
    V1DeploymentSpec,
    V1DeploymentStatus,
    V1DeploymentStrategy,
    V1DownwardAPIProjection,
    V1DownwardAPIVolumeFile,
    V1DownwardAPIVolumeSource,
    V1EmptyDirVolumeSource,
    V1Endpoint,
    V1EndpointAddress,
    V1EndpointConditions,
    V1EndpointHints,
    V1Endpoints,
    V1EndpointSlice,
    V1EndpointSliceList,
    V1EndpointsList,
    V1EndpointSubset,
    V1EnvFromSource,
    V1EnvVar,
    V1EnvVarSource,
    V1EphemeralContainer,
    V1EphemeralVolumeSource,
    V1EventSource,
    V1Eviction,
    V1ExecAction,
    V1ExternalDocumentation,
    V1FCVolumeSource,
    V1FlexPersistentVolumeSource,
    V1FlexVolumeSource,
    V1FlockerVolumeSource,
    V1ForZone,
    V1GCEPersistentDiskVolumeSource,
    V1GitRepoVolumeSource,
    V1GlusterfsPersistentVolumeSource,
    V1GlusterfsVolumeSource,
    V1GroupVersionForDiscovery,
    V1GRPCAction,
    V1HorizontalPodAutoscaler,
    V1HorizontalPodAutoscalerList,
    V1HorizontalPodAutoscalerSpec,
    V1HorizontalPodAutoscalerStatus,
    V1HostAlias,
    V1HostPathVolumeSource,
    V1HTTPGetAction,
    V1HTTPHeader,
    V1HTTPIngressPath,
    V1HTTPIngressRuleValue,
    V1Ingress,
    V1IngressBackend,
    V1IngressClass,
    V1IngressClassList,
    V1IngressClassParametersReference,
    V1IngressClassSpec,
    V1IngressList,
    V1IngressLoadBalancerIngress,
    V1IngressLoadBalancerStatus,
    V1IngressPortStatus,
    V1IngressRule,
    V1IngressServiceBackend,
    V1IngressSpec,
    V1IngressStatus,
    V1IngressTLS,
    V1IPBlock,
    V1ISCSIPersistentVolumeSource,
    V1ISCSIVolumeSource,
    V1Job,
    V1JobCondition,
    V1JobList,
    V1JobSpec,
    V1JobStatus,
    V1JobTemplateSpec,
    V1JSONSchemaProps,
    V1KeyToPath,
    V1LabelSelector,
    V1LabelSelectorRequirement,
    V1Lease,
    V1LeaseList,
    V1LeaseSpec,
    V1Lifecycle,
    V1LifecycleHandler,
    V1LimitRange,
    V1LimitRangeItem,
    V1LimitRangeList,
    V1LimitRangeSpec,
    V1ListMeta,
    V1LoadBalancerIngress,
    V1LoadBalancerStatus,
    V1LocalObjectReference,
    V1LocalSubjectAccessReview,
    V1LocalVolumeSource,
    V1ManagedFieldsEntry,
    V1MutatingWebhook,
    V1MutatingWebhookConfiguration,
    V1MutatingWebhookConfigurationList,
    V1Namespace,
    V1NamespaceCondition,
    V1NamespaceList,
    V1NamespaceSpec,
    V1NamespaceStatus,
    V1NetworkPolicy,
    V1NetworkPolicyEgressRule,
    V1NetworkPolicyIngressRule,
    V1NetworkPolicyList,
    V1NetworkPolicyPeer,
    V1NetworkPolicyPort,
    V1NetworkPolicySpec,
    V1NetworkPolicyStatus,
    V1NFSVolumeSource,
    V1Node,
    V1NodeAddress,
    V1NodeAffinity,
    V1NodeCondition,
    V1NodeConfigSource,
    V1NodeConfigStatus,
    V1NodeDaemonEndpoints,
    V1NodeList,
    V1NodeSelector,
    V1NodeSelectorRequirement,
    V1NodeSelectorTerm,
    V1NodeSpec,
    V1NodeStatus,
    V1NodeSystemInfo,
    V1NonResourceAttributes,
    V1NonResourceRule,
    V1ObjectFieldSelector,
    V1ObjectMeta,
    V1ObjectReference,
    V1Overhead,
    V1OwnerReference,
    V1PersistentVolume,
    V1PersistentVolumeClaim,
    V1PersistentVolumeClaimCondition,
    V1PersistentVolumeClaimList,
    V1PersistentVolumeClaimSpec,
    V1PersistentVolumeClaimStatus,
    V1PersistentVolumeClaimTemplate,
    V1PersistentVolumeClaimVolumeSource,
    V1PersistentVolumeList,
    V1PersistentVolumeSpec,
    V1PersistentVolumeStatus,
    V1PhotonPersistentDiskVolumeSource,
    V1Pod,
    V1PodAffinity,
    V1PodAffinityTerm,
    V1PodAntiAffinity,
    V1PodCondition,
    V1PodDisruptionBudget,
    V1PodDisruptionBudgetList,
    V1PodDisruptionBudgetSpec,
    V1PodDisruptionBudgetStatus,
    V1PodDNSConfig,
    V1PodDNSConfigOption,
    V1PodFailurePolicy,
    V1PodFailurePolicyOnExitCodesRequirement,
    V1PodFailurePolicyOnPodConditionsPattern,
    V1PodFailurePolicyRule,
    V1PodIP,
    V1PodList,
    V1PodOS,
    V1PodReadinessGate,
    V1PodResourceClaim,
    V1PodSchedulingGate,
    V1PodSecurityContext,
    V1PodSpec,
    V1PodStatus,
    V1PodTemplate,
    V1PodTemplateList,
    V1PodTemplateSpec,
    V1PolicyRule,
    V1PortStatus,
    V1PortworxVolumeSource,
    V1Preconditions,
    V1PreferredSchedulingTerm,
    V1PriorityClass,
    V1PriorityClassList,
    V1Probe,
    V1ProjectedVolumeSource,
    V1QuobyteVolumeSource,
    V1RBDPersistentVolumeSource,
    V1RBDVolumeSource,
    V1ReplicaSet,
    V1ReplicaSetCondition,
    V1ReplicaSetList,
    V1ReplicaSetSpec,
    V1ReplicaSetStatus,
    V1ReplicationController,
    V1ReplicationControllerCondition,
    V1ReplicationControllerList,
    V1ReplicationControllerSpec,
    V1ReplicationControllerStatus,
    V1ResourceAttributes,
    V1ResourceClaim,
    V1ResourceFieldSelector,
    V1ResourceQuota,
    V1ResourceQuotaList,
    V1ResourceQuotaSpec,
    V1ResourceQuotaStatus,
    V1ResourceRequirements,
    V1ResourceRule,
    V1Role,
    V1RoleBinding,
    V1RoleBindingList,
    V1RoleList,
    V1RoleRef,
    V1RollingUpdateDaemonSet,
    V1RollingUpdateDeployment,
    V1RollingUpdateStatefulSetStrategy,
    V1RuleWithOperations,
    V1RuntimeClass,
    V1RuntimeClassList,
    V1Scale,
    V1ScaleIOPersistentVolumeSource,
    V1ScaleIOVolumeSource,
    V1ScaleSpec,
    V1ScaleStatus,
    V1Scheduling,
    V1ScopedResourceSelectorRequirement,
    V1ScopeSelector,
    V1SeccompProfile,
    V1Secret,
    V1SecretEnvSource,
    V1SecretKeySelector,
    V1SecretList,
    V1SecretProjection,
    V1SecretReference,
    V1SecretVolumeSource,
    V1SecurityContext,
    V1SelfSubjectAccessReview,
    V1SelfSubjectAccessReviewSpec,
    V1SelfSubjectRulesReview,
    V1SelfSubjectRulesReviewSpec,
    V1SELinuxOptions,
    V1ServerAddressByClientCIDR,
    V1Service,
    V1ServiceAccount,
    V1ServiceAccountList,
    V1ServiceAccountTokenProjection,
    V1ServiceBackendPort,
    V1ServiceList,
    V1ServicePort,
    V1ServiceSpec,
    V1ServiceStatus,
    V1SessionAffinityConfig,
    V1StatefulSet,
    V1StatefulSetCondition,
    V1StatefulSetList,
    V1StatefulSetOrdinals,
    V1StatefulSetPersistentVolumeClaimRetentionPolicy,
    V1StatefulSetSpec,
    V1StatefulSetStatus,
    V1StatefulSetUpdateStrategy,
    V1Status,
    V1StatusCause,
    V1StatusDetails,
    V1StorageClass,
    V1StorageClassList,
    V1StorageOSPersistentVolumeSource,
    V1StorageOSVolumeSource,
    V1Subject,
    V1SubjectAccessReview,
    V1SubjectAccessReviewSpec,
    V1SubjectAccessReviewStatus,
    V1SubjectRulesReviewStatus,
    V1Sysctl,
    V1Taint,
    V1TCPSocketAction,
    V1TokenRequestSpec,
    V1TokenRequestStatus,
    V1TokenReview,
    V1TokenReviewSpec,
    V1TokenReviewStatus,
    V1Toleration,
    V1TopologySelectorLabelRequirement,
    V1TopologySelectorTerm,
    V1TopologySpreadConstraint,
    V1TypedLocalObjectReference,
    V1TypedObjectReference,
    V1UncountedTerminatedPods,
    V1UserInfo,
    V1ValidatingWebhook,
    V1ValidatingWebhookConfiguration,
    V1ValidatingWebhookConfigurationList,
    V1Volume,
    V1VolumeAttachment,
    V1VolumeAttachmentList,
    V1VolumeAttachmentSource,
    V1VolumeAttachmentSpec,
    V1VolumeAttachmentStatus,
    V1VolumeDevice,
    V1VolumeError,
    V1VolumeMount,
    V1VolumeNodeAffinity,
    V1VolumeNodeResources,
    V1VolumeProjection,
    V1VsphereVirtualDiskVolumeSource,
    V1WebhookConversion,
    V1WeightedPodAffinityTerm,
    V1WindowsSecurityContextOptions,
    V2ContainerResourceMetricSource,
    V2ContainerResourceMetricStatus,
    V2CrossVersionObjectReference,
    V2ExternalMetricSource,
    V2ExternalMetricStatus,
    V2HorizontalPodAutoscaler,
    V2HorizontalPodAutoscalerBehavior,
    V2HorizontalPodAutoscalerCondition,
    V2HorizontalPodAutoscalerList,
    V2HorizontalPodAutoscalerSpec,
    V2HorizontalPodAutoscalerStatus,
    V2HPAScalingPolicy,
    V2HPAScalingRules,
    V2MetricIdentifier,
    V2MetricSpec,
    V2MetricStatus,
    V2MetricTarget,
    V2MetricValueStatus,
    V2ObjectMetricSource,
    V2ObjectMetricStatus,
    V2PodsMetricSource,
    V2PodsMetricStatus,
    V2ResourceMetricSource,
    V2ResourceMetricStatus,
)



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Callable, Generic, Iterator, Optional, Type, TypeVar, cast, overload

from kubernetes_dynamic.events import Event, Watch
from kubernetes_dynamic.exceptions import EventTimeoutError
from kubernetes_dynamic.formatters import format_selector
from kubernetes_dynamic.kube.exceptions import ConflictError, NotFoundError, UnprocessibleEntityError
from kubernetes_dynamic.models.resource_value import ResourceValue

R = TypeVar("R", bound=ResourceValue)

if TYPE_CHECKING:
    from kubernetes_dynamic.client import K8sClient
    from kubernetes_dynamic.models.common import ItemList
    from kubernetes_dynamic.models.resource_item import CheckResult


class _Missing:
    pass


MISSING = _Missing()


class ResourceApi(Generic[R]):
    """Represents an API resource type, containing the information required to build urls for requests"""

    def __init__(
        self,
        prefix=None,
        group=None,
        api_version=None,
        kind=None,
        namespaced=False,
        verbs=None,
        name=None,
        preferred=False,
        client=None,
        singularName=None,
        shortNames=None,
        categories=None,
        subresources=None,
        resource_type=None,
        **kwargs,
    ):
        if None in (api_version, kind, prefix):
            raise ValueError("At least prefix, kind, and api_version must be provided")

        self.prefix = prefix
        self.group = group
        self.api_version = api_version
        self.kind = kind
        self.namespaced = namespaced
        self.verbs = verbs
        self.name = name
        self.preferred = preferred
        self.singular_name = singularName or (name[:-1] if name else "")
        self.short_names = shortNames
        self.categories = categories
        self.subresources = {k: Subresource(self, **v) for k, v in (subresources or {}).items()}
        self.client: K8sClient = client
        self.resource_type: Optional[Type[R]] = resource_type

        self.extra_args = kwargs

    def to_dict(self):
        d = {
            "_type": "Resource",
            "prefix": self.prefix,
            "group": self.group,
            "api_version": self.api_version,
            "kind": self.kind,
            "namespaced": self.namespaced,
            "verbs": self.verbs,
            "name": self.name,
            "preferred": self.preferred,
            "singularName": self.singular_name,
            "shortNames": self.short_names,
            "categories": self.categories,
            "subresources": {k: sr.to_dict() for k, sr in self.subresources.items()},
        }
        d.update(self.extra_args)
        return d

    @property
    def group_version(self) -> str:
        if self.group:
            return "{}/{}".format(self.group, self.api_version)
        return cast(str, self.api_version)

    def __repr__(self):
        return "<{}({}/{})>".format(self.__class__.__name__, self.group_version, self.name)

    @property
    def urls(self) -> dict[str, str]:
        full_prefix = "{}/{}".format(self.prefix, self.group_version)
        resource_name = self.name.lower()
        return {
            "base": "/{}/{}".format(full_prefix, resource_name),
            "namespaced_base": "/{}/namespaces/{{namespace}}/{}".format(full_prefix, resource_name),
            "full": "/{}/{}/{{name}}".format(full_prefix, resource_name),
            "namespaced_full": "/{}/namespaces/{{namespace}}/{}/{{name}}".format(full_prefix, resource_name),
        }

    def path(self, name=None, namespace=None) -> str:
        url_type = []
        path_params = {}
        if self.namespaced and namespace:
            url_type.append("namespaced")
            path_params["namespace"] = namespace
        if name:
            url_type.append("full")
            path_params["name"] = name
        else:
            url_type.append("base")
        return self.urls["_".join(url_type)].format(**path_params)

    def __getattr__(self, name) -> Subresource:
        if name in self.subresources:
            return self.subresources[name]
        raise AttributeError(name)

    def serialize_body(self, body):
        """Serialize body to raw dict so apiserver can handle it

        :param body: kubernetes resource body, current support: Union[Dict, ResourceValue]
        """
        if callable(getattr(body, "to_dict", None)):
            return body.to_dict()
        return body or {}

    def ensure_namespace_param(self, namespace: Optional[str | _Missing], body=None, allow_all=False) -> Optional[str]:
        if not self.namespaced:
            return None
        if namespace is MISSING:
            if body:
                namespace = body.get("metadata", {}).get("namespace", self.config.namespace)
            else:
                namespace = self.client.config.namespace
        if not allow_all and not namespace:
            raise ValueError("Namespace is required for {}.{}".format(self.group_version, self.kind))
        return cast(Optional[str], namespace)

    def ensure_name_param(self, name, body=None) -> str:
        if not name and body:
            name = body.get("metadata", {}).get("name")
        if not name:
            raise ValueError("Name is required for {}.{}".format(self.group_version, self.kind))
        return name

    @overload
    def read(
        self,
        *,
        namespace: Optional[str | _Missing] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        **kwargs,
    ) -> ItemList[R]:
        """Get a list of models.
        Setting namespace to `None` or `""` will yield from all namespaces.

        Args:
            namespace: namespace of the resource. Defaults to `current` set to None/Empty for all.
            label_selector: Label selectors to use for query.
            field_selector: Field selectors to use for query.
        """
        ...  # pragma: no cover

    @overload
    def read(self, name: str, namespace: Optional[str | _Missing] = None, **kwargs) -> R:
        """Get a resource model.

        Args:
            name: name of the resource you want to query.
            namespace: namespace of the resource. Defaults to `current`.

        Raises:
            NotFoundError: When the model does not exists.
        """
        ...  # pragma: no cover

    def read(
        self, name: Optional[str] = None, namespace: Optional[str | _Missing] = MISSING, **kwargs
    ) -> R | ItemList[R]:
        namespace = self.ensure_namespace_param(namespace, allow_all=not name)
        path = self.path(name=name, namespace=namespace)
        kwargs.setdefault("serializer", self.resource_type)
        return self.client.request("get", path, **kwargs)

    @overload
    def get(
        self,
        *,
        namespace: Optional[str | _Missing] = MISSING,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        **kwargs,
    ) -> ItemList[R]:
        """Get a list of models.
        Setting namespace to `None` or `""` will yield from all namespaces.

        Args:
            namespace: namespace of the resource. Defaults to `current` set to None/Empty for all.
            label_selector: Label selectors to use for query.
            field_selector: Field selectors to use for query.
        """
        ...  # pragma: no cover

    @overload
    def get(self, name: str, namespace: Optional[str | _Missing] = MISSING, **kwargs) -> Optional[R]:
        """Get a resource model.

        Args:
            name: name of the resource you want to query.
            namespace: namespace of the resource. Defaults to `current`.
        """
        ...  # pragma: no cover

    def get(
        self, name: Optional[str] = None, namespace: Optional[str | _Missing] = MISSING, **kwargs
    ) -> Optional[R] | ItemList[R]:
        try:
            return self.read(name, namespace, **kwargs) if name else self.read(namespace=namespace, **kwargs)
        except NotFoundError:
            return None

    def find(self, pattern: str, namespace: Optional[str | _Missing] = MISSING, **kwargs) -> list[R]:
        """Find a list of models.
        Setting namespace to `None` or `""` will yield from all namespaces.

        Args:
            pattern: regex pattern for metadata.name.
            namespace: namespace of the resource. Defaults to `current` set to None/Empty for all.
            label_selector: Label selectors to use for query.
            field_selector: Field selectors to use for query.
        """
        items = []
        data: Optional[ItemList[R]] = self.get(namespace=namespace, **kwargs)
        if not data:
            return items
        for item in data:
            if re.match(pattern, item.metadata.name):
                items.append(item)
        return items

    def create(self, body: dict | R, namespace: Optional[str | _Missing] = MISSING, **kwargs) -> R:
        body = self.serialize_body(body)
        namespace = self.ensure_namespace_param(namespace, body)
        path = self.path(namespace=namespace)
        kwargs.setdefault("serializer", self.resource_type)
        return self.client.request("post", path, body=body, **kwargs)

    @overload
    def delete(
        self, name: str, namespace: Optional[str | _Missing] = MISSING, body: Optional[dict | R] = None, **kwargs
    ) -> R:
        ...  # pragma: no cover

    @overload
    def delete(
        self,
        *,
        namespace: Optional[str | _Missing] = MISSING,
        body: Optional[dict | R] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        **kwargs,
    ) -> ItemList[R]:
        ...  # pragma: no cover

    def delete(
        self,
        name: Optional[str] = None,
        namespace: Optional[str | _Missing] = MISSING,
        body: Optional[dict | R] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        **kwargs,
    ) -> ItemList[R] | R:
        if not (name or label_selector or field_selector):
            raise ValueError("At least one of name|label_selector|field_selector is required")
        if self.namespaced and not (label_selector or field_selector):
            namespace = self.ensure_namespace_param(namespace, allow_all=not name)
        path = self.path(name=name, namespace=namespace)
        kwargs.setdefault("serializer", self.resource_type)
        return self.client.request(
            "delete", path, body=body, label_selector=label_selector, field_selector=field_selector, **kwargs
        )

    def replace(
        self, body: dict | R, name: Optional[str] = None, namespace: Optional[str | _Missing] = MISSING, **kwargs
    ) -> R:
        body = self.serialize_body(body)
        name = self.ensure_name_param(name, body)
        namespace = self.ensure_namespace_param(namespace, body)
        path = self.path(name=name, namespace=namespace)
        kwargs.setdefault("serializer", self.resource_type)
        return self.client.request("put", path, body=body, **kwargs)

    def patch(
        self, body: dict | R, name: Optional[str] = None, namespace: Optional[str | _Missing] = MISSING, **kwargs
    ) -> R:
        body = self.serialize_body(body)
        name = self.ensure_name_param(name, body)
        namespace = self.ensure_namespace_param(namespace, body)

        content_type = kwargs.pop("content_type", "application/strategic-merge-patch+json")
        path = self.path(name=name, namespace=namespace)
        kwargs.setdefault("serializer", self.resource_type)
        return self.client.request("patch", path, body=body, content_type=content_type, **kwargs)

    def server_side_apply(
        self,
        body: dict | R,
        name: Optional[str] = None,
        namespace: Optional[str | _Missing] = MISSING,
        force_conflicts: Optional[bool] = None,
        **kwargs,
    ) -> R:
        body = self.serialize_body(body)
        name = self.ensure_name_param(name, body)
        namespace = self.ensure_namespace_param(namespace, body)

        kwargs.update({"content_type": "application/apply-patch+yaml"})
        path = self.path(name=name, namespace=namespace)
        kwargs.setdefault("serializer", self.resource_type)
        return self.client.request("patch", path, body=body, force_conflicts=force_conflicts, **kwargs)

    def watch(
        self,
        namespace: Optional[str | _Missing] = MISSING,
        name: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        resource_version: Optional[str] = None,
        timeout: Optional[float] = None,
        watcher: Optional[Watch] = None,
    ) -> Iterator[Event[R]]:
        namespace = self.ensure_namespace_param(namespace, allow_all=True)
        if name:
            field_selector = field_selector or ""
            field_selector += f",metadata.name={name}"
        watcher = watcher or Watch(self.client.client, self.resource_type)
        if watcher and not resource_version:
            resource_version = watcher.resource_version
        return watcher.stream(
            self.get,
            namespace=namespace or self.client.config.namespace,
            name=None,
            field_selector=field_selector,
            label_selector=label_selector,
            resource_version=resource_version,
            serialize=False,
            timeout_seconds=timeout,
        )

    def wait_until(
        self,
        *,
        namespace: Optional[str | _Missing] = MISSING,
        name: Optional[str] = None,
        check: Callable[[Event], CheckResult],
        field_selector: Optional[str] = None,
        label_selector: Optional[str] = None,
        timeout: int = 30,
        **kwargs,
    ) -> Event:
        """Wait until a certain custom check returns true for a resource returned by the stream."""
        namespace = self.ensure_namespace_param(namespace)
        field_selectors = [] if not field_selector else [format_selector(field_selector)]
        if name:
            field_selectors.append(f"metadata.name={name}")

        last = None
        result = None
        for event in self.watch(
            field_selector=format_selector(field_selectors),
            label_selector=format_selector(label_selector),
            timeout=timeout,
            namespace=namespace,
            **kwargs,
        ):
            last = event
            result = check(event)
            if result:
                return event
        if last is None:
            raise EventTimeoutError(f"Timed out waiting for check. {self.kind} {name} not found.", last=last)
        if result is not None and result.message:
            raise EventTimeoutError(result.message, last=last)
        raise EventTimeoutError(f"Timed out waiting for check on {self.kind} {name} .", last=last)

    def apply(self, body: dict | R, namespace: Optional[str | _Missing] = MISSING, **kwargs) -> R:
        namespace = self.ensure_namespace_param(namespace)
        body["metadata"].setdefault("annotations", {})
        name = body["metadata"]["name"]
        try:
            return self.create(body=body, namespace=namespace, **kwargs)
        except ConflictError:
            pass
        try:
            return self.patch(name=name, body=body, namespace=namespace, **kwargs)
        except UnprocessibleEntityError:
            pass

        self.delete(name=name, namespace=namespace, **kwargs)
        return self.create(body=body, namespace=namespace, **kwargs)


class ResourceList(ResourceApi):
    """Represents a list of API objects"""

    def __init__(self, client, group="", api_version="v1", base_kind="", kind=None, base_resource_lookup=None):
        self.client = client
        self.group = group
        self.api_version = api_version
        self.kind = kind or "{}List".format(base_kind)
        self.base_kind = base_kind
        self.base_resource_lookup = base_resource_lookup
        self.__base_resource = None

    def base_resource(self):
        if self.__base_resource:
            return self.__base_resource
        elif self.base_resource_lookup:
            self.__base_resource = self.client.resources.get(**self.base_resource_lookup)
            return self.__base_resource
        elif self.base_kind:
            self.__base_resource = self.client.resources.get(
                group=self.group, api_version=self.api_version, kind=self.base_kind
            )
            return self.__base_resource
        return None

    def _items_to_resources(self, body):
        """Takes a List body and return a dictionary with the following structure:
        {
            'api_version': str,
            'kind': str,
            'items': [{
                'resource': Resource,
                'name': str,
                'namespace': str,
            }]
        }
        """
        if body is None:
            raise ValueError("You must provide a body when calling methods on a ResourceList")

        api_version = body["apiVersion"]
        kind = body["kind"]
        items = body.get("items")
        if not items:
            raise ValueError("The `items` field in the body must be populated when calling methods on a ResourceList")

        if self.kind != kind:
            raise ValueError(
                "Methods on a {} must be called with a body containing the same kind. Receieved {} instead".format(
                    self.kind, kind
                )
            )

        return {"api_version": api_version, "kind": kind, "items": [self._item_to_resource(item) for item in items]}

    def _item_to_resource(self, item):
        metadata = item.get("metadata", {})
        resource = self.base_resource()
        if not resource:
            api_version = item.get("apiVersion", self.api_version)
            kind = item.get("kind", self.base_kind)
            resource = self.client.resources.get(api_version=api_version, kind=kind)
        return {
            "resource": resource,
            "definition": item,
            "name": metadata.get("name"),
            "namespace": metadata.get("namespace"),
        }

    def to_dict(self):
        return {
            "_type": "ResourceList",
            "group": self.group,
            "api_version": self.api_version,
            "kind": self.kind,
            "base_kind": self.base_kind,
        }

    def __getattr__(self, name):
        if self.base_resource():
            return getattr(self.base_resource(), name)
        return None


class Subresource(ResourceApi):
    """Represents a subresource of an API resource. This generally includes operations
    like scale, as well as status objects for an instantiated resource
    """

    def __init__(self, parent, **kwargs):
        self.parent = parent
        self.prefix = parent.prefix
        self.group = parent.group
        self.api_version = parent.api_version
        self.kind = kwargs.pop("kind")
        self.name = kwargs.pop("name")
        self.subresource = kwargs.pop("subresource", None) or self.name.split("/")[1]
        self.namespaced = kwargs.pop("namespaced", False)
        self.verbs = kwargs.pop("verbs", None)
        self.extra_args = kwargs

    @property
    def urls(self):
        full_prefix = "{}/{}".format(self.prefix, self.group_version)
        return {
            "full": "/{}/{}/{{name}}/{}".format(full_prefix, self.parent.name, self.subresource),
            "namespaced_full": "/{}/namespaces/{{namespace}}/{}/{{name}}/{}".format(
                full_prefix, self.parent.name, self.subresource
            ),
        }

    def to_dict(self):
        d = {
            "kind": self.kind,
            "name": self.name,
            "subresource": self.subresource,
            "namespaced": self.namespaced,
            "verbs": self.verbs,
        }
        d.update(self.extra_args)
        return d

    def __getattr__(self, name) -> Subresource:
        return getattr(self.parent, name)



from __future__ import annotations

import re
import typing
from typing import Callable, Optional

import pydantic
from typing_extensions import Self

from kubernetes_dynamic.events import Event, EventType

from ..exceptions import NotFoundError
from .common import V1ObjectMeta
from .resource_value import ResourceValue

if typing.TYPE_CHECKING:
    from kubernetes_dynamic.kube.resource_api import ResourceApi

    from ..client import K8sClient


class CheckResult:
    """Result object to represent the result and message of a check."""

    def __init__(self, state: bool, message: Optional[str] = None) -> None:
        self.state = state
        self.message = message

    def __bool__(self):
        return bool(self.state)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, CheckResult):
            return self.state == __o.state and self.message == __o.message
        return bool(self) == __o


class ResourceItem(ResourceValue):
    """Kubernetes resource objects."""

    apiVersion: str = ""
    kind: str = ""
    metadata: V1ObjectMeta = pydantic.Field(default_factory=V1ObjectMeta)
    status: ResourceValue = pydantic.Field(default_factory=ResourceValue)

    _client_instance: Optional[K8sClient] = pydantic.PrivateAttr()
    _api_instance: Optional[ResourceApi] = pydantic.PrivateAttr()

    def __init__(
        self,
        definition: dict | ResourceValue | None = None,
        client=None,
        **kwargs,
    ):
        final_def = ResourceItem.merge_definition_with_kwargs(definition, **kwargs)
        defaults = self.get_defaults()
        kind = final_def.get("kind") or self.kind or defaults.get("kind")
        if kind:
            final_def["kind"] = kind
        api_version = final_def.get("apiVersion") or self.apiVersion or defaults.get("apiVersion")
        if api_version:
            final_def["apiVersion"] = api_version

        super().__init__(**final_def)

        self._api_instance: Optional[ResourceApi] = None
        self._client_instance = client

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.metadata.name}, namespace={self.metadata.namespace})"

    @classmethod
    def get_defaults(cls):
        """Get some default values based on the class."""
        values = {}
        for cl in [cls] + list(cls.__bases__):
            parts = re.sub("([a-z0-9])([A-Z])", r"\1 \2", cl.__name__).split()
            match = [part for part in parts if re.match(r"^V\d.*", part)]
            if match:
                version = match[0]
                version_index = parts.index(version) + 1
                values["kind"] = "".join(parts[version_index:])
                values["apiVersion"] = version.lower()
                break
        return values

    @property
    def _api(self) -> ResourceApi:
        """Get resource api."""
        if (
            not self._api_instance
            or self._api_instance.api_version != self.apiVersion
            or self._api_instance.kind != self.kind
        ):
            self._api_instance = self._client.get_api(kind=self.kind, api_version=self.apiVersion)
        return self._api_instance

    @property
    def _client(self) -> K8sClient:
        """Get resource api."""
        if not self._client_instance:
            self._client_instance = self.default_client()
        return self._client_instance

    @classmethod
    def default_client(cls) -> K8sClient:
        """Create a default K8sClient."""
        from ..client import K8sClient

        return K8sClient()

    def refresh_(self) -> Self:
        """Refreshes the local instance with kubernetes data."""
        data = self.read_()
        if data is None:
            raise NotFoundError(self)
        return self._update_attrs(data)

    def patch_(self) -> Self:
        """Updates the Kubernetes resource."""
        return self._update_attrs(self._api.patch(name=self.metadata.name, body=self))

    def create_(self) -> Self:
        """Creates the object in Kubernetes."""
        return self._update_attrs(self._api.create(self))

    def read_(self) -> Optional[Self]:
        """Reads the object in Kubernetes."""
        return self._api.get(self.metadata.name, self.metadata.namespace)

    def delete_(self) -> Self:
        """Deletes the object from Kubernetes."""
        return self._api.delete(self.metadata.name, self.metadata.namespace)

    @staticmethod
    def check_object_conditions(item: ResourceItem) -> CheckResult:
        """Check object conditions."""
        name = item.metadata.name
        if not item.status or not item.status.conditions:
            return CheckResult(False, f"No conditions found on {item.kind} {name}.")
        for condition in item.status.conditions:  # type: ignore
            if condition.status != "True":
                untrue_condition = condition
                break
        else:
            return CheckResult(True, f"All conditions true on {item.kind} {name}.")
        return CheckResult(False, f"Condition not true on {item.kind} {name} : {untrue_condition}.")

    @staticmethod
    def check_replicas_ready(item: ResourceItem) -> CheckResult:
        """Check object replicas."""
        item_dict = item.dict()
        replicas = item_dict["status"].get("replicas", 0)
        name = item.metadata.name
        ready_replicas = item.status.readyReplicas or 0
        if ready_replicas == replicas:
            return CheckResult(True, f"All {replicas} replicas ready on {item.kind} {name}.")
        else:
            return CheckResult(
                False,
                f"{item.kind} {name} only have {ready_replicas} ready replicas out of required {replicas}.",
            )

    @staticmethod
    def check_object_is_ready(item: ResourceItem) -> CheckResult:
        """Checks whether given resource object is ready or not.

        For resources that control replicas it checks whether all replicas are ready, otherwise it
        checks whether all status conditions are true.
        """
        if not item.status:
            return CheckResult(False, "Object has no status description.")
        if item.status.replicas:
            return ResourceItem.check_replicas_ready(item)
        elif item.status.conditions:
            return ResourceItem.check_object_conditions(item)
        else:
            raise NotImplementedError(f"Unimplemented resource readiness check for {item.kind}  {item.metadata.name}.")

    def is_ready(self, refresh: bool = False) -> CheckResult:
        """Check if the resource data successfully satisfies the ready state."""
        if refresh:
            self.refresh_()
        return self.check_object_is_ready(self)

    def wait_until_status(self, status: str, timeout: int = 30) -> Event:
        """Wait until the resource is in a specific status."""

        def status_check(event: Event) -> CheckResult:
            result = event.raw_object["status"]["phase"] == status and event.type != EventType.DELETED
            if result:
                message = f"{self.kind} {self.metadata.name} is {status}."
            else:
                message = f"Last status of {self.kind} {self.metadata.name}: {event.raw_object['status']['phase']}"
            return CheckResult(result, message)

        return self.wait_until(check=status_check, timeout=timeout)

    def wait_until_ready(
        self,
        timeout: int = 30,
        **kwargs,
    ):
        """Waiting until resource is ready."""
        return self.wait_until(
            check=lambda event: self.check_object_is_ready(event.object),
            timeout=timeout,
            **kwargs,
        )

    def wait_until_not_ready(
        self,
        timeout: int = 30,
        **kwargs,
    ) -> Event:
        """Waiting until resource is not ready."""

        def not_ready(event: Event) -> CheckResult:
            result = self.check_object_is_ready(event.object)
            return CheckResult(not result, result.message)

        return self.wait_until(
            check=not_ready,
            timeout=timeout,
            **kwargs,
        )

    def wait_until(self, check: Callable[[Event], CheckResult], timeout: int = 30, **kwargs) -> Event:
        return self.api.wait_until(
            check=check,
            timeout=timeout,
            **kwargs,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from kubernetes_dynamic.formatters import format_selector

from .common import ItemList, get_default
from .resource_item import ResourceItem

if TYPE_CHECKING:
    from .all import V1StatefulSetSpec, V1StatefulSetStatus
    from .pod import V1Pod


class V1StatefulSet(ResourceItem):
    spec: V1StatefulSetSpec = Field(default_factory=lambda: get_default("V1StatefulSetSpec"))
    status: V1StatefulSetStatus = Field(default_factory=lambda: get_default("V1StatefulSetStatus"))

    def get_pods(self) -> ItemList[V1Pod]:
        label_selector = self.spec.selector.matchLabels
        return self._client.pods.get(label_selector=format_selector(label_selector))



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, cast

from pydantic import Field

from kubernetes_dynamic.models.common import V1ObjectMeta, get_default

from ..exceptions import ConflictError
from .resource_item import ResourceItem

if TYPE_CHECKING:
    from .all import V1NamespaceSpec, V1NamespaceStatus


class V1Namespace(ResourceItem):
    spec: V1NamespaceSpec = Field(default_factory=lambda: get_default("V1NamespaceSpec"))
    status: V1NamespaceStatus = Field(default_factory=lambda: get_default("V1NamespaceStatus"))

    def annotate(self, annotations_dict: Dict[str, str]) -> V1Namespace:
        """Add annotation(s) to the namespace."""
        self.refresh_()
        existing: Dict[str, str] = self.metadata.annotations or {}
        existing.update(annotations_dict)
        return self.patch_()

    @classmethod
    def ensure(cls, name: str) -> V1Namespace:
        item = V1Namespace(metadata=cast(V1ObjectMeta, {"name": name}))
        try:
            return item.create_()
        except ConflictError:
            return item.refresh_()
